from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
import os
import setuptools
import sys

__version__ = "1.0.9"


class get_pybind_include(object):
    """Helper class to determine the pybind11 include path
    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked. """

    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        import pybind11
        return pybind11.get_include(self.user)


ext_modules = [
    Extension(
        "texterrors_align",
        ["libs/texterrors_align.cc", "libs/stringvector.cc"],
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
            "libs/",
        ],
        language="c++",
    )
]


def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    import tempfile

    with tempfile.NamedTemporaryFile("w", suffix=".cpp") as f:
        f.write("int main (int argc, char **argv) { return 0; }")
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except setuptools.distutils.errors.CompileError:
            return False
    return True

class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""

    c_opts = {"msvc": ["/EHsc"], "unix": []}

    if sys.platform == "darwin":
        c_opts["unix"] += ["-stdlib=libc++", "-mmacosx-version-min=10.7"]

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        if ct == "unix":
            opts.append('-DVERSION_INFO="%s"' % self.distribution.get_version())
            opts.append('-std=c++17')
            if has_flag(self.compiler, "-fvisibility=hidden"):
                opts.append("-fvisibility=hidden")
        elif ct == "msvc":
            opts.append('/DVERSION_INFO=\\"%s\\"' % self.distribution.get_version())
            opts.append('/std:c++17')
        for ext in self.extensions:
            ext.extra_compile_args = opts
        build_ext.build_extensions(self)


base_dir = os.path.dirname(os.path.realpath(__file__))
def get_requires():
    req_path = os.path.join(base_dir, 'requirements.txt')
    install_requires = open(req_path).read().splitlines()
    return install_requires


with open(os.path.join(base_dir, "README.md")) as fh:
    long_description = fh.read()


setup(
    name="texterrors",
    version=__version__,
    author="Rudolf A Braun",
    author_email="rab014@gmail.com",
    packages=["texterrors"],
    license="Apache-2.0 License",
    url='https://github.com/RuABraun/texterrors',
    description="For WER",
    long_description=long_description,
    long_description_content_type="text/markdown",
    ext_modules=ext_modules,
    cmdclass={"build_ext": BuildExt},
    entry_points={'console_scripts': ['texterrors=texterrors.texterrors:cli']},
    install_requires=get_requires(),
    setup_requires=['pybind11'],
    python_requires='>=3.6'
)
